#!/usr/bin/env python
"""
Merge VCFs stored in google buckets pointed to by `input_keys`.
Output to `output_key` in the same bucket.

merge_vcfs.py "fc-9169fcd1-92ce-4d60-9d2d-d19fd326ff10" "combined.vcf.gz" "consent1/HVH_phs000993_TOPMed_WGS_freeze.8.chr7.hg38.vcf.gz,phg001280.v1.TOPMed_WGS_Amish_v4.genotype-calls-vcf.WGS_markerset_grc38.c2.HMB-IRB-MDS/Amish_phs000956_TOPMed_WGS_freeze.8.chr7.hg38.vcf.gz"
"""
import argparse

from xsamtools import vcf

parser = argparse.ArgumentParser()
parser.add_argument("bucket", type=str, help="Name of GS bucket")
parser.add_argument("output_key", type=str, help="GS bucket location of output file.")
parser.add_argument("input_keys", help="Comma delimited list of GS bucket locations of input files.")
args = parser.parse_args()

keys = args.input_keys.split(",")
print(args.bucket)
print(args.output_key)
for key in keys:
    print(key)
vcf.combine(args.bucket, keys, args.bucket, args.output_key)




import os
import time
from uuid import uuid4
from contextlib import AbstractContextManager
from multiprocessing import Process

import gs_chunked_io as gscio

from terra_notebook_utils import gs


class BlobReaderProcess(AbstractContextManager):
    def __init__(self, bucket_name, key, filepath=None):
        self.filepath = filepath or f"/tmp/{uuid4()}"
        os.mkfifo(self.filepath)
        self.proc = Process(target=self.run, args=(bucket_name, key, self.filepath))
        self.proc.start()
        self._closed = False

    def run(self, bucket_name, key, filepath):
        fd = os.open(filepath, os.O_WRONLY)
        blob = gs.get_client().bucket(bucket_name).get_blob(key)
        blob_reader = gscio.AsyncReader(blob, chunks_to_buffer=1)
        while True:
            data = bytearray(blob_reader.read(blob_reader.chunk_size))
            if not data:
                break
            while data:
                try:
                    k = os.write(fd, data)
                    data = data[k:]
                except BrokenPipeError:
                    time.sleep(1)

    def close(self):
        if not self._closed:
            self._closed = True
            self.proc.join(1)
            os.unlink(self.filepath)
            if not self.proc.exitcode:
                self.proc.terminate()

    def __exit__(self, *args, **kwargs):
        self.close()


class BlobWriterProcess(AbstractContextManager):
    def __init__(self, bucket_name, key, filepath=None):
        self.filepath = filepath or f"/tmp/{uuid4()}"
        os.mkfifo(self.filepath)
        self.proc = Process(target=self.run, args=(bucket_name, key, self.filepath))
        self.proc.start()
        self._closed = False

    def run(self, bucket_name, key, filepath):
        bucket = gs.get_client().bucket(bucket_name)
        with gscio.Writer(key, bucket) as blob_writer:
            with open(filepath, "rb") as fh:
                while True:
                    data = fh.read(blob_writer.chunk_size)
                    if not data:
                        break
                    blob_writer.write(data)

    def close(self):
        if not self._closed:
            self._closed = True
            self.proc.join(300)  # TODO: Consider inter-process communication to know when it's safe to close
            os.unlink(self.filepath)
            if not self.proc.exitcode:
                self.proc.terminate()

    def __exit__(self, *args, **kwargs):
        self.close()



import os
import sys
import typing
import subprocess

paths: typing.Dict[str, str] = dict(htsfile=None, bcftools=None)

def _run(cmd: list, **kwargs):
    p = subprocess.run(cmd, **kwargs)
    p.check_returncode()
    return p

def _samtools_binary_path(name):
    roots = [
        os.path.abspath(os.path.dirname(sys.executable)),  # typical bin path in a virtual env
        "/home/jupyter-user/.local/bin"  # terra installation path
    ]
    for root in roots:
        path = os.path.join(root, name)
        try:
            _run([path, "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return path
        except (FileNotFoundError, subprocess.CalledProcessError):
            pass
    return None

for name in paths:
    paths[name] = _samtools_binary_path(name)
    if paths[name] is None:
        print(f"WARNING: {name} unavailable: htslib or bcftools build failed during installation. "
              "          try `pip install -v xsamtools` to diagnose the problem")

available = {name: paths[name] is not None for name in paths}



import io
from uuid import uuid4
from multiprocessing import cpu_count
import subprocess

from terra_notebook_utils import xprofile

from xsamtools import pipes, vcf, samtools


cores_available = cpu_count()


def _merge(input_filepaths, output_filepath):
    subprocess.run([samtools.paths['bcftools'],
                    "merge",
                    "--no-index",
                    "-o", output_filepath,
                    "-O", "z",
                    "--threads", f"{2 * cores_available}"]
                   + [fp for fp in input_filepaths])


@xprofile.profile("combine")
def combine(src_bucket_name, src_keys, dst_bucket_name, dst_key):
    readers = [pipes.BlobReaderProcess(src_bucket_name, key) for key in src_keys]
    writer = pipes.BlobWriterProcess(dst_bucket_name, dst_key)
    try:
        _merge([r.filepath for r in readers], writer.filepath)
    except Exception:
        raise
    finally:
        for reader in readers:
            reader.close()
        writer.close()



import os
import glob
import tempfile
import subprocess
import traceback
from setuptools import setup, find_packages
from setuptools.command import install, build_py, sdist


install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]


def _run(cmd: list, **kwargs):
    p = subprocess.run(cmd, **kwargs)
    p.check_returncode()
    return p


class BuildPy(build_py.build_py):
    def run(self):
        super().run()
        if not self.dry_run:
            try:
                _run(["tar", "xjf", "htslib.tar.bz2", "-C", "build"])
                _run(["tar", "xjf", "bcftools.tar.bz2", "-C", "build"])
                _run(["make"], cwd="build/htslib")
                _run(["make"], cwd="build/bcftools")
            except subprocess.CalledProcessError:
                print("Failed to build htslib/bcftools:")
                traceback.print_exc()


class Install(install.install):
    def run(self):
        super().run()
        if not self.dry_run:
            try:
                root = os.path.dirname(os.path.abspath(__file__))
                bindir = os.path.join(root, os.path.abspath(self.install_scripts))
                libdir = os.path.join(root, os.path.abspath(self.install_lib))
                includedir = os.path.join(root, os.path.abspath(self.install_headers))
                datadir = os.path.join(root, os.path.abspath(self.install_data))
                _run(["make",
                      f"bindir={bindir}",
                      f"includedir={includedir}",
                      f"libdir={libdir}",
                      f"libexecdir={libdir}",
                      f"datarootdir={datadir}",
                      "INSTALL_MAN=:",
                      "install"], cwd="build/htslib")
                _run(["make",
                      f"bindir={bindir}",
                      f"libdir={libdir}",
                      f"libexecdir={libdir}",
                      "INSTALL_MAN=:",
                      "install"], cwd="build/bcftools")
            except subprocess.CalledProcessError:
                print("Failed to package htslib/bcftools")
                traceback.print_exc()


setup(
    name='xsamtools',
    version='0.1.0',
    description='Lightly modified versions of htslib and bcftools.',
    url='https://github.com/xbrianh/xsamtools.git',
    author='Brian Hannafious',
    author_email='bhannafi@ucsc.edu',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    scripts=glob.glob('scripts/*'),
    zip_safe=False,
    install_requires=install_requires,
    platforms=['MacOS X', 'Posix'],
    test_suite='test',
    cmdclass=dict(install=Install, build_py=BuildPy),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7'
    ]
)
